﻿##### Scriptinfo ##################################################################################
# Scriptreihe:      parse-ActiveSyncDebugLog
# Datum:            2020-10-22
# Version:          V1.01
# Programmierer:    Stephan Walther (WS IT-Solutions)
###################################################################################################

# Funktionen 
    function verbinde-MX {
        <#
        .Notes
           Scriptreihe:      verbinde MX (Exchange Server 2013,2016,2019)
           Datum:            2020-05-24
           Version:          V1.00
           Programmierer:    Stephan Walther (WS IT-Solutions)
 
        .Synopsis
           Die Funktion verbindet die aktuelle PowerShell mit einem Exchange Server.
 
        .DESCRIPTION
           Mit dieser Funktion kann die aktuelle PowerShell-Session bequem mit einem verfuegbaren
           Exchange Server verbunden werden. Dies funktioniert mit den Versionen zwischen 2013 und
           2019. Hybrid-Szenarien und Exchange Online wurden noch nicht getestet. Die Funktion
           ermittelt via LDAP die existierenden Exchange Server und probiert dabei alle durch,
           bis eine erfolgreiche Verbindung zustande kommt oder keine weiteren Server mehr vorhanden
           sind.
 
        .EXAMPLE
           verbinde-MX
 
           Ohne Parameter sucht die Funktion nach einem verfuegbaren Exchange Server und stellt eine
           Verbindung her. Dabei werden Meldungen ausgegeben.
 
        .EXAMPLE
           verbinde-MX -NewSession
 
           Wurde bereits eine Verbindung aufgebaut, dann wird ein erneuter Aufruf dies erkennen und
           keine neue Session erstellen. Mit dem Zusatz -NewSession wird die bestehende Verbindung
           entfernt und neu aufgebaut.
 
        .EXAMPLE
           if (verbinde-MX -ReturnBool -Silent) {
               # erfolgreiche Verbindung
           } else {
               # fehlgeschlagene Verbindung
           }
 
           Diese Variante eignet sich besonders für Bedingungen, da keine Meldungen erzeugt werden.
           Stattdessen wird ein Boolean-Wert für die Bedingung zurueckgegeben. 
      
         .PARAMETER Silent
           Der Parameter unterdrueckt die Meldungstexte. Im Standard werden via Write-Host 
           Meldungen angezeigt.
 
         .PARAMETER ReturnBool
           Der Parameter gibt als Ergebnis des Verbindungaufbaus ein true oder false zurueck. 
           Damit kann die Funktion direkt in if-Bedingungen verwendet werden. Im Standard
           wird nur der Meldungstext ausgegeben.
 
         .PARAMETER NewSession
           Mit diesem Parameter wird eine neue Verbindung erzwungen, wenn schon eine zu
           einem Exchange Server besteht. Die alte Verbindung wird beendet.
        #>
        [outputtype([boolean])]
        param(
            [parameter(Mandatory=$false)] [switch] $Silent     = $false,
            [parameter(Mandatory=$false)] [switch] $ReturnBool = $false,
            [parameter(Mandatory=$false)] [switch] $NewSession = $false
        )

        #region Variablen 
            $weiter = $true
            $WriteHost = -not $silent
        #endregion
    
        #region besteht bereits eine Verbindung?
            if ($weiter) {
                $MXSessions = Get-PSSession | 
                    Where-Object {$_.ConfigurationName -eq 'microsoft.exchange' -and $_.State -eq 'opened'}
                    
                if ( $MXSessions ) {
                    if ( $NewSession ) {
                        try {
                            $MXSessions | Remove-PSSession 
                            if ($WriteHost) {Write-Host "Eine bestehende Verbindung zum Server $($MXSessions.ComputerName -join ' und ') wurde beendet." -ForegroundColor Green}
                        }
                        catch {
                            if ($WriteHost) {Write-Host "FEHLER bei Beenden der Verbindung: $($error[0].Exception.Message)" -ForegroundColor Yellow}
                            $weiter = $false
                        }
                    } else {
                        if ($WriteHost) {Write-Host "Es besteht bereits eine Verbindung zum Server $($MXSessions.ComputerName -join ' und ')." -ForegroundColor Green}
                        $weiter = $false
                    }
                } 
            }
        #endregion
    
        #region Exchange-Powershell-URLs finden
            if ($weiter) {
                try {
                    $URLs = @()

                    # Variante 1: direkte Abfrage im LDAP ==> setzt Exchange-Berechtigung voraus
                        $Partition = "LDAP://" +([adsi] "LDAP://RootDSE").Get("configurationNamingContext")
                        $Searcher = New-Object DirectoryServices.DirectorySearcher
                        $Searcher.SearchRoot = $Partition
                        $Searcher.Filter = "(objectClass=msExchPowerShellVirtualDirectory)"
                        $Searcher.FindAll().Properties.msexchinternalhostname |
                            ForEach-Object {
                                if ($_.length -gt 0) { 
                                    $URLs += $_ 
                                    if ($WriteHost) {Write-Host "Server gefunden: $_" -ForegroundColor Green}
                                }
                            }

                    # Variante 2: falls keine URLs wegen fehlenden Rechten gefunden wurden ==> MX-Server ermitteln
                        if ($URLs.count -eq 0) {
                            $Searcher.Filter = "(objectCategory=msExchExchangeServer)"
                            $Searcher.FindAll().Properties.cn | 
                                ForEach-Object {
                                    $URLs += "http://$_/powershell"
                                    if ($WriteHost) {Write-Host "Server gefunden: http://$_/powershell" -ForegroundColor Green}
                                }
                        } 

                    # Ergebnis:
                        if ($URLs.count -eq 0) {
                            if ($WriteHost) {Write-Host "Es wurden keine Exchange Server gefunden!" -ForegroundColor Yellow}
                            $weiter = $false
                        }            
                }
                catch {
                    if ($WriteHost) {Write-Host "FEHLER bei Verbindung mit LDAP: $($error[0].Exception.Message)" -ForegroundColor Yellow}
                    $weiter = $false
                }
            }
        #endregion
    
        #region Verbindung aufbauen
            if ($weiter) {
                $verbunden = $false
                $URLs | ForEach-Object {
                    if ($verbunden -eq $false){
                        try {
                            $URL = $_
                            $MX = New-PSSession -ConnectionUri $URL -Authentication Kerberos -ConfigurationName microsoft.exchange -ErrorAction Stop
                            Import-PSSession -Session $MX -DisableNameChecking | Out-Null
                            $verbunden = $true 
                            if ($WriteHost) {Write-Host "verbunden mit $URL" -ForegroundColor Green}
                        }
                        catch {
                            if ($WriteHost) {Write-Host "FEHLER bei Verbindung mit '$($URL)': $($error[0].Exception.Message)" -ForegroundColor Yellow}
                        }
                    }
                }    
            }
        #endregion

        #region besteht jetzt eine Verbindung? ==> Ausgabe
            if ($ReturnBool) {
                $MXSessions = Get-PSSession | 
                    Where-Object {$_.ConfigurationName -eq 'microsoft.exchange' -and $_.State -eq 'opened'}
                    
                if ( $MXSessions ) {
                    return $true
                } else {
                    return $false
                }
            }
        #endregion
    }

    function parse-ActiveSyncDebugLog {
        <#
        .Notes
           Scriptreihe:      parse-ActiveSyncDebugLog (Exchange Server 2013,2016,2019)
           Datum:            2020-10-22
           Version:          V1.01
           Programmierer:    Stephan Walther (WS IT-Solutions)
 
        .Synopsis
           Die Funktion kann die ActiveSyncDebuglogs einer Mailbox auslesen und parsen.
 
        .DESCRIPTION
           Mit dieser Funktion kann von einer Mailbox für ein oder mehrere auswählbare ActiveSync-Device 
           das ActiveSyncDebuglog ausgelesen werden. Normalerweise liegen diese Daten in einem Fließtext
           vor. Die Funktion parst die Informationen aus dem Text heraus und gibt sie als PowerShell-
           Objekt strukturiert zurück. 

           Wichtig: Für das Auslesen des Logs sind 2 Voraussetzungen zu erfüllen:
           1) Das ActiveSyncDebugLog muss aktiviert sein.
           2) Nach der Aktivierung muss ein Synchronisierungsversuch vom Device erfolgen, der dann 
              protokolliert wird

           Das DebugLog muss pro Mailbox mit diesem Aufruf aktiviert werden:

               Set-CASMailbox -Identity "Tessa.Test" -ActiveSyncDebugLogging $true

           Nach 72 Stunden deaktiviert der Exchange Server das Logging für die Mailbox automatisch. Nach
           der Analyse sollte es aber manuell mit Set-CASMailbox wieder deaktiviert werden.
 
        .EXAMPLE
           parse-ActiveSyncDebugLog -Mailbox "Tessa.Test"
 
           Für die Angegebene Mailbox werden die ActiveSync-Devices ermittel. Wenn nur eines vorhanden
           ist, dann wird das ActiveSyncDebugLog für dieses Device ausgelesen. Sind mehrere Devices
           registriert, wird vorab eine Liste zur Auswahl angezeigt.

        .EXAMPLE
           parse-ActiveSyncDebugLog -Mailbox "Tessa.Test" -AllDevices
 
           Die Auswahl der Devices wird unterdrueckt. Es werden alle Devices ausgewertet.
            
        .EXAMPLE
           parse-ActiveSyncDebugLog -Mailbox "Tessa.Test" -DeviceID "155117C2317744E297C336FC48425F3A"
 
           Mit der Angabe einer gültigen DeviceID wird automatisch das ActiveSyncDebugLog für dieses 
           Device angezeigt.
 
        .EXAMPLE
           parse-ActiveSyncDebugLog -Mailbox "Tessa.Test" -output gridview
 
           Diese Variante gibt die Ergebnisse des Parsings in einem Gridview aus. Im Gridview koennen
           weitere Filter angewendet werden.

        .EXAMPLE
           parse-ActiveSyncDebugLog -Mailbox "Tessa.Test" -output FormatTable
 
           Diese Variante gibt die Ergebnisse des Parsings in einem FormatTable als Text aus. Einige
           unwichtige Spalten werde hierbei nicht angezeigt.

        .EXAMPLE
           parse-ActiveSyncDebugLog -Mailbox "Tessa.Test" -output HTML
 
           Diese Variante gibt die Ergebnisse des Parsings einer HTML-Datei aus und oeffnet diese im
           Browser.
                            
        .PARAMETER Mailbox
           Der Parameter wird zur Identifikation der Mailbox vom ActiveSync-Device verwendet. Gültige
           Eingaben sind z.B. der Alias, der SamAccountName, der UPN. Wildcards sind nicht zulässig.
 
        .PARAMETER DeviceID
           Der Parameter ist optional. Wird er nicht angegeben, dann wird bei mehreren Geräten eine
           Auswahl durch ein GridView angezeigt.

        .PARAMETER AllDevices
           Dieser Parameter ueberschreibt die Auswahl der ActiveSync-Devices. Es werden die Logs fuer
           alle Devices angezeigt.
            
        .PARAMETER output
           Mit diesem Parameter kann das Ausgabeverhalten der geparsten Ergebnisse gesteuert werden.
           1) OBJECT ist der Standard. Mit diesem Argument wird ein PSObject ausgegeben. Dieses kann 
              mit der Pipeline weiter verarbeitet werden.
           2) GRIDVIEW gibt die Daten als filterbares GridView zurück. Eine Weiterverarbeitung ist
              nicht vorgesehen.
           3) FORMATTABLE gibt eine tabellarische Ansicht aus. Diese enthält relevante Spalten und 
              ist nicht weiter verarbeitbar.                        
           4) HTML gibt eine tabellarische Ansicht in einer HTML-Datei aus.
        #>
        param(
            [Parameter(Mandatory=$true)]                                                          [string]   $Mailbox,
            [Parameter(Mandatory=$false)]                                                         [string[]] $DeviceID,
            [Parameter(Mandatory=$false)]                                                         [switch]   $AllDevices = $false,
            [Parameter(Mandatory=$false)] [ValidateSet('object','gridview','FormatTable','html')] [string]   $output = "object"
        )

        #region Vorprüfung
            try {
                Get-Mailbox -Identity $Mailbox -ErrorAction Stop | Out-Null
                Write-Verbose "Mailbox '$Mailbox' ist vorhanden."
            }
            catch {
                Write-Host "Mailbox '$Mailbox' wurde nicht gefunden!" -ForegroundColor Yellow
                break
            }
        #endregion

        #region Deviceermittlung
            if ($AllDevices -eq $true) {
                Write-Verbose "suche alle ActiveSyncDevices"
                $DeviceID = (Get-MobileDevice -Mailbox $Mailbox).deviceID
            } else {
                if ($PSBoundParameters.Keys -notcontains "DeviceID") {
                    $Device = @()
                    $Device += Get-MobileDevice -Mailbox $Mailbox | 
                        Select-Object -Property Friendlyname,DeviceID,DeviceImei,DeviceOS,DeviceUserAgent,DeviceType
           
                    if ($Device.count -gt 1) { 
                        write-verbose "$($Device.count) Devices gefunden ... zeige Auswahl an"                    
                        $Device = $Device | Out-GridView -PassThru -Title "Wähle ein oder mehrere ActiveSync-Devices aus:"
                    } elseif ($Device.count -eq 1) {
                        write-verbose "nur ein Device gefunden"
                    } else {
                        Write-Host "Fuer die Mailbox '$Mailbox' sind keine ActiveSync-Devices registriert!" -ForegroundColor Yellow
                        break
                    }

                    if ($Device -ne $null) {
                        $DeviceID = $Device.deviceID
                    }        
                } else {
                    $DeviceID | ForEach-Object {
                        $CurID = $_
                        if ((Get-MobileDevice -Mailbox $Mailbox).DeviceID -notcontains $CurID) {
                            Write-Host "Die DeviceID '$CurID' wurde fuer die Mailbox '$Mailbox' nicht gefunden!" -ForegroundColor Yellow
                            break
                        }    
                    }
                }
            }
            Write-Verbose "Anzahl der gewaehlten Devices: $($DeviceID.count)"
        #endregion

        #region Auslesen der DebugLogs
            if ($DeviceID.Length -gt 0) {     
                Write-Verbose "Starte Auslesen der ActiveSyncDebugLogs fuer die Mailbox '$Mailbox' und die DeviceIDs '$( $DeviceID -join ',' )'"
                $ActiveSyncLogs = Get-MobileDeviceStatistics -Mailbox $Mailbox -GetMailboxLog | Where-Object { $DeviceID -contains $_.deviceID }
                
                $LogReport = $ActiveSyncLogs | Select-Object -Property DeviceID,MailboxLogReport,DeviceFriendlyName,DeviceModell,DeviceOS,Status
            }

            $LogLength = ($LogReport | ForEach-Object {$_.MailboxLogReport.Length} | Measure-Object -Sum).sum
        #endregion
        
        #region Parsen der DebugLogs
            if ($LogLength -gt 0) {
                Write-Verbose "Starte das Parsen der Logfiles"
                $Result = @()

                $LogReport | ForEach-Object {
                    $CurLog = $_

                    $CurLog.MailboxLogReport -split "Log Entry: " | Select-Object -Skip 1 | ForEach-Object {
                        $CurRecord = $_
                        $LogEntry = ($CurRecord -split "`r`n")[0].trim()
                        Write-Verbose "parse LogEntry $LogEntry"
    
                        $AllProperties = @()
                        $AllProperties += "RequestTime :"
                        $AllProperties += "ServerName :"
                        $AllProperties += "AssemblyVersion :"
                        $AllProperties += "Identifier :"
                        $AllProperties += "RequestHeader :"
                        $AllProperties += "RequestBody :"
                        $AllProperties += "LogicalRequest :"
                        $AllProperties += "AccessState :"
                        $AllProperties += "AccessStateReason :"
                        $AllProperties += "ResponseHeader :"
                        $AllProperties += "ResponseBody :"
                        $AllProperties += "ResponseTime :"

                        $CurProperties = @()
                        $AllProperties | ForEach-Object {
                            if ($CurRecord | Select-String $_) { $CurProperties += $_ }
                        }
    
                        $PropertyPattern = $CurProperties -join ' |'

                        $Values = $CurRecord -split $PropertyPattern

                        $i = 1
                        $Ausgabe = New-Object -TypeName psobject
                        $Ausgabe | Add-Member -MemberType NoteProperty -Name DeviceID           -Value $CurLog.DeviceID
                        $Ausgabe | Add-Member -MemberType NoteProperty -Name DeviceFriendlyName -Value $CurLog.DeviceFriendlyName
                        $Ausgabe | Add-Member -MemberType NoteProperty -Name DeviceModell       -Value $CurLog.DeviceModell
                        $Ausgabe | Add-Member -MemberType NoteProperty -Name DeviceOS           -Value $CurLog.DeviceOS
                        $Ausgabe | Add-Member -MemberType NoteProperty -Name Status             -Value $CurLog.Status
                        $Ausgabe | Add-Member -MemberType NoteProperty -Name LogEntry           -Value $LogEntry

                        $CurProperties | ForEach-Object {
                            $CurProperty = $_ -replace ' :'
                            $CurValue = $Values[$i].trim()

                            if ('RequestTime','ResponseTime' -contains $CurProperty) { 
                                $CurTimeStamp = ($CurValue -split "`r`n")[0]
                                $CurValue = Get-Date -Date ([DateTime] $CurTimeStamp) -Format 'yyyy-MM-dd HH:mm:ss'
                            }
                
                            if ('RequestHeader' -contains $CurProperty) {
                                $CurValue -split "`r`n" | ForEach-Object {
                                    $Parts = $_ -split ':'
                                    $SubProperty  = $Parts[0]
                                    $SubValue = (($Parts | Select-Object -Skip 1) -join ':').trim()

                                    if ($SubProperty -like 'POST /Microsoft-Server-ActiveSync*') {
                                        $SubValue = $SubProperty
                                        $SubProperty = "URI"
                                    }

                                    $Ausgabe | Add-Member -MemberType NoteProperty -Name $SubProperty -Value $SubValue
                                }
                            } else {
                                if ('ResponseHeader' -contains $CurProperty) { $CurValue = ($CurValue -split "`r`n")[0] }

                                $Ausgabe | Add-Member -MemberType NoteProperty -Name $CurProperty -Value $CurValue
                            }        
        
                            $i += 1
                        }
                        $Result += $Ausgabe
                    } 
                }
            } else {
                Write-Host "Keine Logs fuer Mailbox '$Mailbox' und die Devices '$( $DeviceID -join ',' )' gefunden! Ist das DebugLogging aktiv? Versuche `r`n`r`n   Set-CASMailbox -Identity '$Mailbox' -ActiveSyncDebugLogging `$true`r`n`r`nund starte einen neuen Sync vom Device." -ForegroundColor Yellow
            }
        #endregion

        #region Ausgabe
            if ($Result.count -gt 0) {
                $FinalResult = $Result | 
                    Select-Object -Property *, @{ n="Latency" ; e={ ([DateTime] $_.ResponseTime - [DateTime] $_.RequestTime).TotalSeconds }}

                if ($output -eq 'gridview') {
                    $FinalResult | 
                        Select-Object -Property * -ExcludeProperty 'URI','Cache-Control','Pragma','X-CommonAccessToken','x-MSExchangeActivityCtx' | 
                            Out-GridView
                } elseif ($output -eq 'FormatTable') {
                    $FinalResult | 
                        Format-Table -AutoSize  -Property DeviceFriendlyName,LogEntry,RequestTime,Latency,ServerName,ResponseHeader,
                                AccessState,AccessStateReason,'X-FE-ClientIP','X-OriginalRequestHost'
                } elseif ($output -eq 'HTML') {
                    $Style = @()
                    $Style += "<style>"
                    $Style += "   TABLE{border: 1px solid black; border-collapse: collapse; font-size: 8pt;}"
                    $Style += "   TH{border: 1px solid black; background: #dddddd; padding: 5px; color: #000000;}"
                    $Style += "   TD{border: 1px solid black; padding: 5px; }"
                    $Style += "</style>"
                    $Style = $Style -join "`r`n"

                    $FinalResult | ConvertTo-Html -Title "ActiveSyncDebugLog fuer Mailbox '$Mailbox'" -Head $Style | Out-File "$( $env:TEMP )\ActiveSyncDebugLog.htm" 
                    & "$( $env:TEMP )\ActiveSyncDebugLog.htm" 
                                
                } else {
                    Return $FinalResult
                }
            }
        #endregion
    }

    break

# Vorbereitung
    verbinde-MX

    Set-CASMailbox -Identity tessa.test -ActiveSyncDebugLogging $true

# Aufrufe
    Get-Help -Name parse-ActiveSyncDebugLog -ShowWindow

    parse-ActiveSyncDebugLog -Mailbox "tessa.test" 
    parse-ActiveSyncDebugLog -Mailbox "tessa.test" -Verbose
    parse-ActiveSyncDebugLog -Mailbox "tessa.test" -output gridview
    parse-ActiveSyncDebugLog -Mailbox "tessa.test" -output FormatTable
    parse-ActiveSyncDebugLog -Mailbox "tessa.test" -output html
    parse-ActiveSyncDebugLog -Mailbox "tessa.test" -AllDevices -output FormatTable

# FailChecks
    parse-ActiveSyncDebugLog -Mailbox "tessa.test" -DeviceID 12345
    parse-ActiveSyncDebugLog -Mailbox "keine.Mailbox" 
    parse-ActiveSyncDebugLog -Mailbox "Paul.Paulsen"   # kein AS-Device